SystemOrganization addCategory: #'TaskScheduler-Core'!

Link subclass: #Task
	instanceVariableNames: 'runTime activity'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'TaskScheduler-Core'!

!Task commentStamp: 'HenrikSperreJohansen 1/23/2012 14:08' prior: 0!
I hold an activity to perform, conforming to the cull: protocol.
The action will receive the elapsed time since last run as the optional parameter.

As the tasks will ideally run in a separate process from the GUI, do NOT force a redraw from tasks.
If you do, you might run into trouble (sooner or later).

Rather, use the changed:/invalidRect: protocol, or #addDeferredUIMessage.
These get scheduled to run at next iteration of the UIProcess loop.!

!Task class methodsFor: 'instance creation' stamp: 'HenrikSperreJohansen 1/23/2012 12:46'!
activity: anActivity
	^self new activity: anActivity! !

!Task methodsFor: 'initialize-release' stamp: 'HenrikSperreJohansen 1/23/2012 12:47'!
activity: anActivity
	activity := anActivity! !

!Task methodsFor: 'dispatch' stamp: 'HenrikSperreJohansen 1/23/2012 16:28'!
addTo: aTaskList
	^aTaskList privateAddTask: self! !

!Task methodsFor: 'performing' stamp: 'HenrikSperreJohansen 1/23/2012 12:48'!
executeAfter: elapsedTime

	runTime := Time millisecondsToRun: [activity cull: elapsedTime]! !

!Task methodsFor: 'initialize-release' stamp: 'HenrikSperreJohansen 1/23/2012 12:42'!
initialize
	runTime := 0.! !

!Task methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 1/23/2012 12:52'!
runTime
	^runTime! !

Announcement subclass: #TaskEjected
	instanceVariableNames: 'task'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'TaskScheduler-Core'!

!TaskEjected commentStamp: 'HenrikSperreJohansen 1/23/2012 13:01' prior: 0!
I get announced if a certain task takes too long to complete, and has to be removed from the scheduled interval.!

Object subclass: #TaskList
	instanceVariableNames: 'delay runTime tasks runLoop'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'TaskScheduler-Core'!

!TaskList commentStamp: 'HenrikSperreJohansen 1/23/2012 12:32' prior: 0!
TaskList contains a list of tasks to complete at a certain interval (in microseconds).

If any of my tasks takes too long, I announce that it is being ejected.
It is up to a Scheduler or similar to decide what to do in that case, f.ex. by logging a warning and insert the item at a lower frequency.!

!TaskList class methodsFor: 'as yet unclassified' stamp: 'HenrikSperreJohansen 1/23/2012 16:58'!
interval: aMillisecondInterval
	^self new interval: aMillisecondInterval! !

!TaskList methodsFor: 'add/remove' stamp: 'HenrikSperreJohansen 1/23/2012 17:02'!
add: aTask
	tasks addLast: aTask.
	runLoop ifNil: [self start].! !

!TaskList methodsFor: 'as yet unclassified' stamp: 'HenrikSperreJohansen 1/23/2012 16:57'!
ejectSlowestTask 
	| slowest |
	slowest := Task new.
	tasks linksDo: [:each | each runTime > slowest runTime ifTrue: [slowest := each]].
	self announce: (TaskEjected task: 
					(tasks removeLink: slowest))! !

!TaskList methodsFor: 'as yet unclassified' stamp: 'HenrikSperreJohansen 1/23/2012 16:56'!
executeAfter: elapsedTime
	|onSchedule|
	onSchedule := interval > Time millisecondsToRun: [tasks linksDo: [:each | each executeAfter: elapsedTime]].
	onSchedule ifFalse: [self ejectSlowestTask]
	! !

!TaskList methodsFor: 'initialize-release' stamp: 'HenrikSperreJohansen 1/23/2012 16:56'!
initialize
	tasks := LinkedList new.! !

!TaskList methodsFor: 'initialize-release' stamp: 'HenrikSperreJohansen 1/23/2012 16:59'!
interval: anInterval
	delay := Delay new: anInterval! !

!TaskList methodsFor: 'run loop' stamp: 'HenrikSperreJohansen 1/23/2012 17:13'!
start
	runLoop := 
		[|start|
		start := Time millisecondClockValue.
		[|end|delay wait. 
		self executeAfter: (end := Time millisecondClockValue) - start.
		start := end.] repeat] 
			newProcess priority: Processor userInterruptPriority. 
	runLoop resume.! !

Object subclass: #TaskScheduler
	instanceVariableNames: 'refreshRate delay taskLists'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'TaskScheduler-Core'!

!TaskScheduler commentStamp: 'HenrikSperreJohansen 1/23/2012 12:38' prior: 0!
I provide an API for scheduling Tasks to run at regular intervals, as well as debugging facilities to help the tasks run smoothly.

I should be run in a separate thread from GUI activites and event handling.!

!TaskScheduler methodsFor: 'initialize-release' stamp: 'HenrikSperreJohansen 1/23/2012 16:53'!
initialize
	taskLists := Set new.
! !

!TaskScheduler methodsFor: 'run loop' stamp: 'HenrikSperreJohansen 1/23/2012 16:40'!
processTasks
	"This is where I run the TaskLists whose interval has been exceeded"! !

!TaskScheduler methodsFor: 'adding' stamp: 'HenrikSperreJohansen 1/23/2012 17:00'!
schedule: aTask every: milliSeconds
	 | list |
	list := taskLists at: milliSeconds ifAbsentPut: [TaskList interval: milliSeconds].
	list add: aTask

	! !

!TaskScheduler methodsFor: 'run loop' stamp: 'HenrikSperreJohansen 1/23/2012 16:39'!
start
	[delay wait.
	self processTasks.
	self updateDelay.

	 ]whileTrue.! !

!TaskScheduler methodsFor: 'run loop' stamp: 'HenrikSperreJohansen 1/23/2012 16:41'!
updateDelay
	"This is where I update the delay to wait an amount of time equal to the shortest remaining"! !
